from __future__ import absolute_import, print_function, unicode_literals

import contextlib

import six
from django import template
from django.template.base import token_kwargs
from django.utils.module_loading import import_string

import django_ftl

register = template.Library()

MODE_SERVER = 'server'
MODES = [
    MODE_SERVER
]

MODE_VAR_NAME = '__ftl_mode'
BUNDLE_VAR_NAME = '__ftl_bundle'


@register.simple_tag(takes_context=True)
def ftlconf(context, mode=None, bundle=None):
    if mode is not None:
        validate_mode(mode)
        context[MODE_VAR_NAME] = mode
    if bundle is not None:
        context[BUNDLE_VAR_NAME] = resolve_bundle(bundle)
    return ''


def resolve_bundle(bundle):
    if isinstance(bundle, six.text_type):
        return import_string(bundle)
    else:
        return bundle


@register.simple_tag(takes_context=True)
def ftlmsg(context, message_id, **kwargs):
    try:
        mode = context[MODE_VAR_NAME]
    except KeyError:
        raise ValueError("No mode set for ftl - use ftlconf/withftl to set mode")
    try:
        bundle = context[BUNDLE_VAR_NAME]
    except KeyError:
        raise ValueError("No bundle set for ftl - use ftlconf/withftl to set bundle")
    if mode == MODE_SERVER:
        return bundle.format(message_id, kwargs)
    raise AssertionError("Not reached")


def validate_mode(mode):
    if mode not in MODES:
        raise ValueError("mode '{0}' not understood, must be one of {1}"
                         .format(mode, MODES))


class WithFtlNode(template.Node):
    def __init__(self, nodelist, language=None, mode=None, bundle=None):
        self.nodelist = nodelist
        self.language = language
        self.mode = mode
        self.bundle = bundle

    def __repr__(self):
        return '<%s>' % self.__class__.__name__

    def render(self, context):
        language = None if self.language is None else self.language.resolve(context)
        mode = None if self.mode is None else self.mode.resolve(context)
        if mode is not None:
            validate_mode(mode)
        bundle = None if self.bundle is None else resolve_bundle(self.bundle.resolve(context))
        new_context = {}
        if mode is not None:
            new_context[MODE_VAR_NAME] = mode
        if bundle is not None:
            new_context[BUNDLE_VAR_NAME] = bundle
        if language is not None:
            lang_ctx = django_ftl.override(language)
        else:
            lang_ctx = null_ctx()

        with context.push(**new_context):
            with lang_ctx:
                return self.nodelist.render(context)


@register.tag('withftl')
def withftl(parser, token):
    conf = token_kwargs(token.split_contents()[1:], parser, support_legacy=False)
    language = conf.pop('language', None)
    mode = conf.pop('mode', None)
    bundle = conf.pop('bundle', None)
    if conf:
        raise ValueError("withftl tag received unexpected keyword arguments: {0}"
                         .format(", ".join(conf.keys())))
    nodelist = parser.parse(('endwithftl',))
    parser.delete_first_token()

    return WithFtlNode(nodelist,
                       language=language,
                       mode=mode,
                       bundle=bundle)


@contextlib.contextmanager
def null_ctx():
    yield



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django.shortcuts import render
from django.template.response import TemplateResponse
from django.test import override_settings
from django.utils.translation import override as dj_override

from django_ftl.bundles import NoLocaleSet

from .base import WebTestBase
from .ftl_bundles import simple_view as simple_view_bundle


def simple_view(request):
    return render(request, 'tests/simple_view.html', {})


def simple_template_response_view(request):
    return TemplateResponse(request, 'tests/simple_view.html', {})


@override_settings(
    MIDDLEWARE=[
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django_ftl.middleware.activate_from_request_session',
    ],
)
class TestActivateFromSessionMiddleware(WebTestBase):
    def test_default(self):
        self.get_url('test_middleware.simple_view')
        self.assertTextPresent("A Web Page Title")
        self.assertTextPresent("The current language code is en")

    def test_set_language(self):
        self.get_url('test_middleware.simple_view')
        self.fill({'[name=language]': 'tr'})
        self.submit('[type=submit]')
        self.assertTextPresent("Şimdiki dil kodu tr")

    def test_request_isolation(self):
        self.get_url('test_middleware.simple_view')
        self.fill({'[name=language]': 'tr'})
        self.submit('[type=submit]')

        self.assertRaises(NoLocaleSet, simple_view_bundle.format, 'simple-title')

    def test_template_response(self):
        # We are checking that the middleware mechanism that finalizes and
        # renders templates works with our use of an 'override' context manager.
        self.get_url('test_middleware.simple_view_template_response')
        self.assertTextPresent("A Web Page Title")
        self.assertTextPresent("The current language code is en")
        self.fill({'[name=language]': 'tr'})
        self.submit('[type=submit]')
        self.assertTextPresent("Şimdiki dil kodu tr")


@override_settings(
    MIDDLEWARE=[
        'django.middleware.locale.LocaleMiddleware',
        'django_ftl.middleware.activate_from_request_language_code',
    ],
)
class TestActivateFromLanguageCodeMiddleware(WebTestBase):
    def test_default(self):
        with dj_override('en'):
            self.get_url('test_middleware.simple_view_prefixed')
        self.assertTextPresent("A Web Page Title")
        self.assertTextPresent("The current language code is en")

    def test_other(self):
        with dj_override('tr'):
            self.get_url('test_middleware.simple_view_prefixed')
        self.assertTextPresent("Şimdiki dil kodu tr")

    def test_request_isolation(self):
        with dj_override('en'):
            self.get_url('test_middleware.simple_view')

        self.assertRaises(NoLocaleSet, simple_view_bundle.format, 'simple-title')

    def test_template_response(self):
        # We are checking that the middleware mechanism that finalizes and
        # renders templates works with our use of an 'override' context manager.
        with dj_override('en'):
            self.get_url('test_middleware.simple_view_template_response_prefixed')
            self.assertTextPresent("A Web Page Title")
            self.assertTextPresent("The current language code is en")
        with dj_override('tr'):
            self.get_url('test_middleware.simple_view_template_response_prefixed')
            self.assertTextPresent("Şimdiki dil kodu tr")



from __future__ import absolute_import, print_function, unicode_literals

from django.conf import settings
from django.utils.translation import LANGUAGE_SESSION_KEY

from django_ftl import override


def activate_from_request_session(get_response):
    """
    Middleware that can be placed after django.middleware.session.SessionMiddleware,
    used in conjunction with `set_language` Django view, and uses
    request.session[LANGUAGE_SESSION_KEY] for Fluent translations.
    """
    def middleware(request):
        language_code = request.session.get(LANGUAGE_SESSION_KEY, settings.LANGUAGE_CODE)
        request.LANGUAGE_CODE = language_code
        with override(language_code):
            return get_response(request)

    return middleware


def activate_from_request_language_code(get_response):
    """
    Middleware that can be placed after django.middleware.locale.LocaleMiddleware,
    and uses the request.LANGUAGE_CODE to activate the same language
    for Fluent translations.

    Requires USE_I18N = True.
    """
    def middleware(request):
        with override(request.LANGUAGE_CODE):
            return get_response(request)

    return middleware



from __future__ import absolute_import, print_function, unicode_literals

from django.utils.decorators import ContextDecorator

from .bundles import activator

__version__ = '0.10'


def activate(locale):
    """
    Activate a given locale/language. Bundles will
    use this locale for translation.
    """
    activator.activate(locale)


def deactivate():
    """
    Deactivate the current locale/language. Bundles will
    fall back to the default locale (if require_activate=False)
    """
    activator.deactivate()


class override(ContextDecorator):
    def __init__(self, locale, deactivate=False):
        self.locale = locale
        self.deactivate = deactivate

    def __enter__(self):
        self.old_locale = activator.get_current_value()
        activator.activate(self.locale)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.deactivate:
            activator.deactivate()
        else:
            activator.activate(self.old_locale)



from __future__ import absolute_import, print_function, unicode_literals

import logging

import django.core.signals as django_signals
import pyinotify

logger = logging.getLogger(__name__)


class BundleModifiedHandler(pyinotify.ProcessEvent):

    def my_init(self, reloader=None):
        self.reloader = reloader

    def process_IN_CLOSE_WRITE(self, evt):
        logger.debug("Clearing bundle due to changed file {0}".format(evt.pathname))
        self.reloader.trigger_reload()


class Reloader(object):
    def __init__(self, bundle):
        self.bundle = bundle
        self.wm = pyinotify.WatchManager()
        self.watches = {}
        self.handler = BundleModifiedHandler(reloader=self)
        # ThreadedNotifier seems to give us horrible problems with Django's
        # devserver when it comes to auto reloading. So we use a Notifier and
        # run the checks for every new request.
        logger.debug("Creating notifier for {0}".format(bundle))
        self.notifier = pyinotify.Notifier(self.wm, self.handler, timeout=100)
        django_signals.request_started.connect(self.new_request)

    def new_request(self, sender, **kwargs):
        try:
            self.notifier.process_events()
            while self.notifier.check_events():  # loop in case more events appear while we are processing
                self.notifier.read_events()
                self.notifier.process_events()
        except RuntimeError:
            # Sometimes get:
            #  RuntimeError: concurrent poll() invocation
            pass

    def add_watched_path(self, path):
        logger.debug("Observing {0} for changes".format(path))
        self.watches.update(self.wm.add_watch(path, pyinotify.IN_CLOSE_WRITE))

    def trigger_reload(self):
        # Remove watches, because the Bundle will add them again after a reload.
        self.wm.rm_watch(list(self.watches.values()))
        self.bundle.reload()


def create_bundle_reloader(bundle):
    return Reloader(bundle)



from __future__ import absolute_import, print_function, unicode_literals

from django.test import override_settings

from django_ftl.conf import get_setting

from .base import TestBase


class TestSettings(TestBase):

    def test_default_fallback(self):
        self.assertEqual(get_setting('LANGUAGE_CODE'), 'en')

    @override_settings(FTL={'LANGUAGE_CODE': 'fr'})
    def test_ftl_override(self):
        self.assertEqual(get_setting('LANGUAGE_CODE'), 'fr')

    @override_settings(FTL={'LANGUAGE_CODE': 'fr'},
                       LANGUAGE_COOKIE_NAME='xxxx')
    def test_ftl_partial_override(self):
        self.assertEqual(get_setting('LANGUAGE_COOKIE_NAME'), 'xxxx')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals


def make_namespace(**attributes):
    class namespace(object):
        pass

    namespace = namespace()
    for k, v in attributes.items():
        setattr(namespace, k, v)

    return namespace



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import re

from django.template import Context, Template

from django_ftl import activate
from django_ftl.bundles import Bundle

from .base import TestBase

main_bundle = Bundle(['tests/main.ftl'],
                     use_isolating=False,
                     default_locale='en')

other_bundle = Bundle(['tests/other.ftl'],
                      use_isolating=False,
                      default_locale='en')


class TestFtlConfTag(TestBase):
    def setUp(self):
        activate('en')

    def test_good(self):
        t = Template("""
        {% load ftl %}
        {% ftlconf mode='server' bundle='tests.test_templatetags.main_bundle' %}
        {% ftlmsg 'simple' %}
        """)
        self.assertEqual(t.render(Context({})).strip(),
                         "Simple")

    def test_good_split(self):
        t = Template("""
        {% load ftl %}
        {% ftlconf mode='server' %}
        {% ftlconf bundle='tests.test_templatetags.main_bundle' %}
        {% ftlmsg 'simple' %}
        """)
        self.assertEqual(t.render(Context({})).strip(),
                         "Simple")

    def test_missing(self):
        t = Template("""
        {% load ftl %}
        {% ftlconf mode='server' bundle='tests.test_templatetags.main_bundle' %}
        {% ftlmsg 'missing-message' %}
        """)
        self.assertEqual(t.render(Context({})).strip(),
                         "???")

    def test_args(self):
        t = Template("""
        {% load ftl %}
        {% ftlconf mode='server' bundle='tests.test_templatetags.main_bundle' %}
        {% ftlmsg 'with-argument' user=user %}
        """)
        self.assertEqual(t.render(Context({'user': 'Mary'})).strip(),
                         "Hello to Mary.")

    def test_xss(self):
        t = Template("""
        {% load ftl %}
        {% ftlconf mode='server' bundle='tests.test_templatetags.main_bundle' %}
        {% ftlmsg 'with-argument' user=user %}
        """)
        self.assertEqual(t.render(Context({'user': 'Mary & Jane'})).strip(),
                         "Hello to Mary &amp; Jane.")


class TestWithFtlTag(TestBase):
    def setUp(self):
        activate('en')

    def test_good(self):
        t = Template("""
        {% load ftl %}
        {% withftl mode='server' bundle='tests.test_templatetags.main_bundle' %}
           {% ftlmsg 'simple' %}
        {% endwithftl %}
        """)
        self.assertEqual(t.render(Context({})).strip(),
                         "Simple")

    def test_override_bundle(self):
        t = Template("""
        {% load ftl %}
        {% ftlconf mode='server' bundle='tests.test_templatetags.main_bundle' %}
        {% ftlmsg 'simple' %}
        {% withftl bundle='tests.test_templatetags.other_bundle' %}
           {% ftlmsg 'other-message' %}
        {% endwithftl %}
        {% ftlmsg 'simple' %}
        """)
        self.assertEqual(re.split(r'\s+', t.render(Context({})).strip()),
                         ["Simple", "Other-message", "Simple"])

    def test_override_language(self):
        t = Template("""
        {% load ftl %}
        {% ftlconf mode='server' bundle='tests.test_templatetags.main_bundle' %}
        {% ftlmsg 'simple' %}
        {% withftl language='fr-FR' %}
           {% ftlmsg 'simple' %}
        {% endwithftl %}
        {% ftlmsg 'simple' %}
        """)
        self.assertEqual(re.split(r'\s+', t.render(Context({})).strip()),
                         ["Simple", "Facile", "Simple"])

    def test_override_bundle_and_language(self):
        t = Template("""
        {% load ftl %}
        {% ftlconf mode='server' bundle='tests.test_templatetags.main_bundle' %}
        {% ftlmsg 'simple' %}
        {% withftl bundle='tests.test_templatetags.other_bundle' language='tr' %}
           {% ftlmsg 'other-message' %}
        {% endwithftl %}
        {% ftlmsg 'simple' %}
        """)
        self.assertEqual(re.split(r'\s+', t.render(Context({})).strip()),
                         ["Simple", "Başka-mesaj", "Simple"])

    def test_bad_kwarg(self):
        t = """
        {% load ftl %}
        {% withftl xyz='abc' %}
           {% ftlmsg 'simple' %}
        {% endwithftl %}
        """
        self.assertRaises(ValueError, Template, t)

    def test_good_with_variables(self):
        t = Template("""
        {% load ftl %}
        {% withftl mode=my_mode bundle=my_bundle %}
           {% ftlmsg 'simple' %}
        {% endwithftl %}
        """)
        context = Context({
            'my_mode': 'server',
            'my_bundle': main_bundle,
        })
        self.assertEqual(t.render(context).strip(),
                         "Simple")

    def test_nesting(self):
        t = Template("""
        {% load ftl %}
        {% withftl mode='server' bundle='tests.test_templatetags.main_bundle' %}
          {% ftlmsg 'simple' %}
          {% withftl language='fr-FR' %}
            {% ftlmsg 'simple' %}
            {% withftl language='tr' %}
              {% ftlmsg 'simple' %}
            {% endwithftl %}
            {% ftlmsg 'simple' %}
          {% endwithftl %}
          {% ftlmsg 'simple' %}
        {% endwithftl %}
        """)
        self.assertEqual(re.split(r'\s+', t.render(Context({})).strip()),
                         ["Simple", "Facile", "Basit", "Facile", "Simple"])

    def test_withftl_bad_mode(self):
        t = Template("""
        {% load ftl %}
        {% withftl mode='xxx' bundle='tests.test_templatetags.main_bundle' %}
          {% ftlmsg 'simple' %}
        {% endwithftl %}
        """)
        self.assertRaises(ValueError, t.render, Context({}))

    def test_ftlconf_bad_mode(self):
        t = Template("""
        {% load ftl %}
        {% ftlconf mode='xxx' bundle='tests.test_templatetags.main_bundle' %}
        {% ftlmsg 'simple' %}
        """)
        self.assertRaises(ValueError, t.render, Context({}))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import six
from django.utils.html import SafeText, mark_safe

from django_ftl.bundles import Bundle

from .base import TestBase

text_type = six.text_type


bundle = Bundle(['tests/escaping.ftl'],
                default_locale='en')


class TestBundles(TestBase):
    maxDiff = None

    def test_html(self):
        val = bundle.format('my-test-item-html', {'name': 'Me & My Friends'})
        self.assertEqual(val, 'Welcome to \u2068Jack &amp; Jill\u2069. \u2068Jack &amp; Jill <i>ROCK</i> - <b>Yeah!</b>\u2069. Your name is \u2068Me &amp; My Friends\u2069.')
        self.assertEqual(type(val), SafeText)

    def test_html_mark_safe(self):
        val = bundle.format('my-test-item-html', {'name': mark_safe('<b>Me</b>')})
        self.assertEqual(val, 'Welcome to \u2068Jack &amp; Jill\u2069. \u2068Jack &amp; Jill <i>ROCK</i> - <b>Yeah!</b>\u2069. Your name is \u2068<b>Me</b>\u2069.')
        self.assertEqual(type(val), SafeText)

    def test_plain(self):
        val = bundle.format('my-test-item-plain', {'name': 'Me & My Friends'})
        self.assertEqual(val, 'Welcome to \u2068Jack & Jill\u2069. Your name is \u2068Me & My Friends\u2069.')
        self.assertEqual(type(val), text_type)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import threading
import time

import six
from django.test import override_settings
from django.utils.encoding import force_text

from django_ftl import activate, deactivate, override
from django_ftl.bundles import Bundle, FileNotFoundError, NoLocaleSet, locale_lookups

from .base import TestBase

text_type = six.text_type


class TestBundles(TestBase):
    def test_no_locale_set_with_require_activate(self):
        bundle = Bundle(['tests/main.ftl'],
                        default_locale='en',
                        require_activate=True)
        self.assertRaises(NoLocaleSet, bundle.format, 'simple')

    def test_require_activate_after_activate(self):
        bundle = Bundle(['tests/main.ftl'],
                        default_locale='en',
                        require_activate=True)
        activate('en')
        self.assertEqual(bundle.format('simple'), 'Simple')
        deactivate()
        self.assertRaises(NoLocaleSet, bundle.format, 'simple')

    def test_no_locale_set_with_default_set(self):
        bundle = Bundle(['tests/main.ftl'],
                        require_activate=True,
                        default_locale='en')
        self.assertRaises(NoLocaleSet, bundle.format, 'simple')

    def test_no_locale_set_with_good_default(self):
        bundle = Bundle(['tests/main.ftl'],
                        default_locale='en')
        self.assertEqual(bundle.format('simple'), 'Simple')

    @override_settings(FTL={'LANGUAGE_CODE': 'en'})
    def test_no_locale_set_with_good_default_from_settings(self):
        bundle = Bundle(['tests/main.ftl'])
        self.assertEqual(bundle.format('simple'), 'Simple')

    @override_settings(FTL={'LANGUAGE_CODE': 'zh'})
    def test_no_locale_set_with_missing_default(self):
        bundle = Bundle(['tests/main.ftl'])
        self.assertRaises(FileNotFoundError, bundle.format, 'simple')

    def test_default_locale_lazy(self):
        # Ensure that bundle is retrieving LANGUAGE_CODE lazily. (The only real
        # reason for this at the moment is to make testing easier).
        bundle = Bundle(['tests/main.ftl'])
        with override_settings(LANGUAGE_CODE='fr-FR'):
            self.assertEqual(bundle.format('simple'), 'Facile')

    def test_find_messages(self):
        bundle = Bundle(['tests/main.ftl'])
        activate('en')
        self.assertEqual(bundle.format('simple'), "Simple")

    def test_load_multiple_with_some_missing(self):
        bundle = Bundle(['tests/only_in_en.ftl',
                         'tests/main.ftl'],
                        default_locale='en')
        activate('fr-FR')
        self.assertEqual(bundle.format('simple'), "Facile")

    def test_switch_locale(self):
        bundle = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle.format('simple'), "Simple")
        activate('tr')
        self.assertEqual(bundle.format('simple'), "Basit")
        deactivate()
        self.assertEqual(bundle.format('simple'), "Simple")

    def test_override(self):
        bundle = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle.format('simple'), "Simple")
        with override('tr'):
            self.assertEqual(bundle.format('simple'), "Basit")
            with override('fr-FR'):
                self.assertEqual(bundle.format('simple'), "Facile")
            self.assertEqual(bundle.format('simple'), "Basit")
        self.assertEqual(bundle.format('simple'), "Simple")

    def test_fallback(self):
        activate('tr')
        bundle = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle.format('missing-from-others'), "Missing from others")

    def test_missing(self):
        activate('en')
        bundle = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle.format('missing-from-all'), "???")

    def test_locale_matching_case_insensitive(self):
        activate('fr-fr')
        bundle = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle.format('simple'), 'Facile')

        activate('EN')
        bundle = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle.format('simple'), 'Simple')

    def test_handle_underscores_in_locale_name(self):
        activate('fr_FR')
        bundle = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle.format('simple'), 'Facile')

    def test_locale_matching_for_default_locale(self):
        activate('zh')
        bundle = Bundle(['tests/main.ftl'], default_locale='EN')  # 'EN' not 'en'
        self.assertEqual(bundle.format('simple'), 'Simple')

    def test_locale_range_lookup(self):
        # en-US does not exist, but 'en' does and should be found
        activate('en-US')
        bundle = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle.format('simple'), 'Simple')

    def test_locale_range_lookup_missing(self):
        # There is no fr or fr-XY (only fr-FR), neither of these should fallback
        # to fr-FR
        activate('fr')
        bundle = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle.format('simple'), 'Simple')

        activate('fr-XY')
        bundle = Bundle(['tests/main.ftl'])
        self.assertEqual(bundle.format('simple'), 'Simple')

    def test_locale_range_lookup_list(self):
        # fr-XY doesn't exist, fr-FR does
        activate('fr-XY, fr-FR')
        bundle = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle.format('simple'), 'Facile')

        # en-GB doesn't exist, en does
        activate('en-GB, en')
        bundle = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle.format('simple'), 'Simple')

    def test_missing_ftl_file(self):
        activate('en')
        bundle = Bundle(['tests/non-existant.ftl'], default_locale='en')
        self.assertRaises(FileNotFoundError, bundle.format, 'simple')

    def test_number_formatting(self):
        bundle = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle.format('with-number-argument', {'points': 1234567}),
                         'Score: \u20681,234,567\u2069')
        activate('fr-FR')
        self.assertEqual(bundle.format('with-number-argument', {'points': 1234567}),
                         'Points: \u20681\xa0234\xa0567\u2069')
        deactivate()
        self.assertEqual(bundle.format('with-number-argument', {'points': 1234567}),
                         'Score: \u20681,234,567\u2069')

    def test_number_formatting_for_fallback(self):
        # When we fall back to a default, number formatting
        # should be consistent with the language.
        # German locale: 1.234.567
        # System locale  (would probably be 'en',): 1,234,567
        # French locale: 1 234 567⁩

        bundle = Bundle(['tests/main.ftl'], default_locale='fr-FR')
        activate('de')
        # Should get French words and formatting
        self.assertEqual(bundle.format('with-number-argument', {'points': 1234567}),
                         'Points: \u20681\xa0234\xa0567\u2069')

    # TODO - check caches are actually working

    def test_lazy(self):
        bundle = Bundle(['tests/main.ftl'], default_locale='en')

        l = bundle.format_lazy('simple')
        self.assertEqual(force_text(l), 'Simple')
        activate('fr-FR')
        self.assertEqual(force_text(l), 'Facile')
        deactivate()
        self.assertEqual(force_text(l), 'Simple')

    def test_lazy_with_require_activate(self):
        bundle = Bundle(['tests/main.ftl'],
                        default_locale='en',
                        require_activate=True)
        self.assertRaises(NoLocaleSet, bundle.format, 'simple')
        msg = bundle.format_lazy('simple')

        self.assertRaises(NoLocaleSet, force_text, msg)

        activate('en')
        self.assertEqual(force_text(msg), 'Simple')
        activate('fr-FR')
        self.assertEqual(force_text(msg), 'Facile')

    def test_prevent_module_level_format(self):
        try:
            import tests.prevent_module_level_format  # noqa
        except NoLocaleSet:
            pass
        else:
            self.fail("Expected NoLocaleSet error")

    def test_allow_module_level_format_lazy(self):
        import tests.allow_module_level_format_lazy
        self.assertRaises(NoLocaleSet,
                          text_type,
                          tests.allow_module_level_format_lazy.MyThing.my_label)
        activate('fr-FR')
        self.assertEqual(force_text(tests.allow_module_level_format_lazy.MyThing.my_label),
                         'Facile')

    def test_use_isolating(self):
        bundle_1 = Bundle(['tests/main.ftl'], default_locale='en')
        self.assertEqual(bundle_1.format('with-argument', {'user': 'Horace'}),
                         'Hello to \u2068Horace\u2069.')

        bundle_2 = Bundle(['tests/main.ftl'], default_locale='en', use_isolating=False)
        self.assertEqual(bundle_2.format('with-argument', {'user': 'Horace'}),
                         'Hello to Horace.')


class TestLocaleLookups(TestBase):
    # See https://tools.ietf.org/html/rfc4647#section-3.4

    def test_language(self):
        self.assertEqual(list(locale_lookups("en")),
                         ["en"])

    def test_language_and_region(self):
        self.assertEqual(list(locale_lookups("en-US")),
                         ["en-US", "en"])

    def test_extra_subtags(self):
        self.assertEqual(list(locale_lookups("zh-Hant-CN-x-private1-private2")),
                         ["zh-Hant-CN-x-private1-private2",
                          "zh-Hant-CN-x-private1",
                          "zh-Hant-CN",
                          "zh-Hant",
                          "zh"])

    def test_list(self):
        self.assertEqual(list(locale_lookups("en-US, fr-FR")),
                         ["en-US", "en", "fr-FR", "fr"])

    def test_eliminates_dupes(self):
        # This would be en-GB, en, en without dupe elimination
        self.assertEqual(list(locale_lookups("en-GB, en")),
                         ["en-GB",
                          "en"])


class TestBundleThreadSafe(TestBase):
    def test_two_threads(self):
        # Not a proof, but a demonstration that a single Bundle can handle
        # threads with different locale values without getting confused.

        bundle = Bundle(['tests/main.ftl'],
                        default_locale='en',
                        require_activate=True)
        lock = threading.Lock()

        output = []

        # Primitive coordination mechanism to ensure we
        # are getting interleaving.
        def wait_until_output(length):
            while True:
                with lock:
                    if len(output) < length:
                        time.sleep(0)
                    else:
                        return

        def thread_1():
            with lock:
                activate('en')
                output.append((1, bundle.format('simple')))
            wait_until_output(2)
            with lock:
                # Should still be in English,
                output.append((1, bundle.format('simple')))

        def thread_2():
            # Make sure thread_1 goes first:
            wait_until_output(1)
            with lock:
                activate('fr-FR')
                output.append((2, bundle.format('simple')))
            wait_until_output(3)
            with lock:
                # Should still be French
                output.append((2, bundle.format('simple')))
                activate('en')
                # Should allow switching
                output.append((2, bundle.format('simple')))

        t1 = threading.Thread(target=thread_1)
        t2 = threading.Thread(target=thread_2)
        t1.start()
        t2.start()
        t1.join()
        t2.join()

        self.assertEqual(output,
                         [
                             (1, 'Simple'),
                             (2, 'Facile'),
                             (1, 'Simple'),
                             (2, 'Facile'),
                             (2, 'Simple'),
                         ])



from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
from collections import OrderedDict
from threading import Lock, local

import six
from django.conf import settings
from django.utils import lru_cache
from django.utils.functional import cached_property, lazy
from django.utils.html import SafeText
from django.utils.html import conditional_escape as conditional_html_escape
from django.utils.html import mark_safe as mark_html_escaped
from fluent.runtime import CompilingFluentBundle as FluentBundle

from .conf import get_setting
from .utils import make_namespace

_active_locale = local()

ftl_logger = logging.getLogger('django_ftl.message_errors')

text_type = six.text_type


class NoLocaleSet(AssertionError):
    pass


def get_app_locale_dirs():
    from django.apps import apps
    dirs = []
    for app_config in apps.get_app_configs():
        if not app_config.path:
            continue
        locales_dir = os.path.join(app_config.path, "locales")
        if os.path.isdir(locales_dir):
            dirs.append(locales_dir)
    return dirs


class FileNotFoundError(RuntimeError):
    pass


class MessageFinderBase(object):

    @property
    def locale_base_dirs(self):
        raise NotImplementedError()

    def load(self, locale, path, reloader=None):
        locale = normalize_bcp47(locale)
        all_bases = self.locale_base_dirs
        tried = []
        for i, base in enumerate(all_bases):
            try:
                locale_dir = locale_dirs_at_path(base)[locale]
            except KeyError:
                tried.append(base + "/")
                continue

            full_path = os.path.join(locale_dir, path)
            try:
                with open(full_path, "rb") as f:
                    if reloader is not None:
                        reloader.add_watched_path(full_path)
                    return f.read().decode('utf-8')

            except (IOError, OSError):
                tried.append(full_path)

        raise FileNotFoundError("Could not find locate FTL file {0}/{1}. Tried: {2}"
                                .format(locale, path,
                                        ", ".join(tried)))


def normalize_bcp47(locale):
    return locale.lower().replace('_', '-')


@lru_cache.lru_cache(maxsize=None)
def locale_dirs_at_path(base):
    dirs_found = [p for p in
                  [os.path.join(base, e) for e in os.listdir(base)]
                  if os.path.isdir(p)]

    # Apply normalization and return mapping from locale nane to directory.
    return {normalize_bcp47(os.path.basename(p)): p
            for p in dirs_found}


class DjangoMessageFinder(MessageFinderBase):

    @cached_property
    def locale_base_dirs(self):
        return get_app_locale_dirs()


default_finder = DjangoMessageFinder()


class LanguageActivator(object):

    def activate(self, locale):
        old_value = self.get_current_value()
        if old_value == locale:
            return
        _active_locale.value = locale

    def deactivate(self):
        self.activate(None)

    def get_current_value(self):
        return getattr(_active_locale, 'value', None)


activator = LanguageActivator()


html_escaper = make_namespace(
    name="django_html_escaper",
    select=lambda message_id=None, **hints: message_id.endswith('-html'),
    output_type=SafeText,
    mark_escaped=mark_html_escaped,
    escape=conditional_html_escape,
    join=lambda parts: mark_html_escaped(''.join(conditional_html_escape(p) for p in parts)),
    use_isolating=True,
)


class Bundle(object):
    def __init__(self, paths,
                 finder=default_finder,
                 default_locale=None,
                 use_isolating=True,
                 require_activate=False,
                 auto_reload=None):

        self._paths = paths
        self._finder = finder
        self._default_locale = default_locale
        self._use_isolating = use_isolating
        self._require_activate = require_activate
        self._lock = Lock()

        if auto_reload is None:
            auto_reload = get_setting('AUTO_RELOAD_BUNDLES', None)

        if auto_reload is None:
            if settings.DEBUG:
                try:
                    import pyinotify  # noqa
                except ImportError:
                    ftl_logger.warning("Not using django_ftl autoreloader because pyinotify is not installed. Set `Bundle.auto_reload` or `AUTO_RELOAD_BUNDLES` to `False` to disable this warning.")
                else:
                    auto_reload = True

        if auto_reload:
            # import at this point to avoid importing pyinotify if we don't need it.
            from .autoreload import create_bundle_reloader
            self._reloader = create_bundle_reloader(self)
        else:
            self._reloader = None
        self.reload()

    def _make_fluent_bundle(self, locale):
        return FluentBundle([locale], use_isolating=self._use_isolating,
                            escapers=[html_escaper])

    def reload(self):
        with self._lock:
            self._all_fluent_bundles = {}  # dict from available locale to FluentBundle
            self._fluent_bundles_for_current_locale = {}  # dict from chosen locale to list of FluentBundle

    def _get_default_locale(self):
        default_locale = self._default_locale
        if default_locale is not None:
            return default_locale

        return get_setting('LANGUAGE_CODE')

    def get_fluent_bundles_for_current_locale(self):
        current_locale = activator.get_current_value()
        if current_locale is None:
            if self._require_activate:
                raise NoLocaleSet("activate() must be used before using Bundle.format "
                                  "- or, use Bundle(require_activate=False)")
        try:
            return self._fluent_bundles_for_current_locale[current_locale]
        except KeyError:
            pass

        # Fill out _fluent_bundles_for_current_locale and _all_fluent_bundles as
        # necessary, but do this synchronized for all threads.
        with self._lock:
            # Double checked locking pattern.
            try:
                return self._fluent_bundles_for_current_locale[current_locale]
            except KeyError:
                pass

            to_try = list(locale_lookups(current_locale)) if current_locale is not None else []
            default_locale = self._get_default_locale()
            if default_locale is not None:
                to_try = uniquify(to_try + [default_locale])

            bundles = []
            for i, locale in enumerate(to_try):
                try:
                    bundle = self._all_fluent_bundles[locale]
                    if bundle is not None:
                        bundles.append(bundle)
                except KeyError:
                    # Create the FluentBundle on demand
                    bundle = self._make_fluent_bundle(locale)

                    for path in self._paths:
                        try:
                            contents = self._finder.load(locale, path, reloader=self._reloader)
                        except FileNotFoundError:
                            if locale == default_locale:
                                # Can't find any FTL with the specified filename, we
                                # want to bail early and alert developer.
                                raise
                            # Allow missing files otherwise
                        else:
                            bundle.add_messages(contents)
                    errors = bundle.check_messages()
                    for msg_id, error in errors:
                        self._log_error(bundle, msg_id, {}, error)
                    bundles.append(bundle)
                    self._all_fluent_bundles[locale] = bundle

            # Shortcut next time
            self._fluent_bundles_for_current_locale[current_locale] = bundles
            return bundles

    def format(self, message_id, args=None):
        fluent_bundles = self.get_fluent_bundles_for_current_locale()
        for i, bundle in enumerate(fluent_bundles):
            try:
                value, errors = bundle.format(message_id, args=args)
                for e in errors:
                    self._log_error(bundle, message_id, args, e)
                return value
            except LookupError as e:
                self._log_error(bundle, message_id, args, e)

        # None were successful, return default
        return '???'

    format_lazy = lazy(format, text_type)

    def _log_error(self,
                   bundle,
                   message_id,
                   args,
                   exception):
        # TODO - nicer error that includes path and source line
        ftl_logger.error("FTL exception for locale [%s], message '%s', args %r: %s",
                         ", ".join(bundle.locales),
                         message_id,
                         args,
                         repr(exception))


def locale_lookups(locale):
    """
    Utility for implementing RFC 4647 lookup algorithm
    """
    if ',' in locale:
        locales = [l.strip() for l in locale.split(",")]
        return uniquify(sum(map(locale_lookups, locales), []))

    parts = locale.split('-')
    locales = []
    current = None
    for p in parts:
        if current is None:
            current = p
            locales.append(current)
        else:
            current = current + "-" + p
            if len(p) == 1:
                # Skip single letter/digit bits
                continue
            locales.append(current)

    return list(reversed(locales))


def uniquify(l):
    return list(OrderedDict.fromkeys(l))



from __future__ import absolute_import, print_function, unicode_literals

from django.conf import settings

NOT_PROVIDED = object()


def get_setting(name, default=NOT_PROVIDED):
    # Look in FTL dict for setting, otherwise fallback to settings module,
    # otherwise fallback to default
    setting_dict = getattr(settings, 'FTL', {})

    if default is NOT_PROVIDED:
        fallback = getattr(settings, name)
    else:
        fallback = getattr(settings, name, default)
    return setting_dict.get(name, fallback)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django_ftl.bundles import Bundle

simple_view = Bundle(['tests/simple_view.ftl'],
                     default_locale='en',
                     use_isolating=False,
                     require_activate=True)



# -*- coding: utf-8
from __future__ import absolute_import, print_function, unicode_literals

DEBUG = True
USE_TZ = True
USE_I18N = True

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "-vgnk4x&72$zmxhp$q6sch83v%@mp()&6*+kf2du0r7l#8i2v-"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

ROOT_URLCONF = "tests.urls"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django_ftl.apps.DjangoFtlConfig",
    "tests",
]

STATIC_URL = '/static/'

SITE_ID = 1

MIDDLEWARE = ()

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                "django.template.context_processors.request",
            ]
        },
    },
]

LANGUAGE_CODE = 'en'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django.test import TestCase
from django_functest import FuncWebTestMixin

from django_ftl import deactivate


class TestBase(TestCase):
    def setUp(self):
        super(TestBase, self).setUp()
        deactivate()


class WebTestBase(FuncWebTestMixin, TestBase):
    setup_auth = False



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django.conf.urls import url
from django.conf.urls.i18n import i18n_patterns
from django.views import i18n as i18n_views

from . import test_middleware

urlpatterns = [
    url(r'^set_language/$', i18n_views.set_language, name='set_language'),
    url(r'^test_middleware/simple_view/$', test_middleware.simple_view, name='test_middleware.simple_view'),
    url(r'^test_middleware/simple_view_template_response/$', test_middleware.simple_view, name='test_middleware.simple_view_template_response'),
] + i18n_patterns(
    url(r'^test_middleware/simple_view/$', test_middleware.simple_view, name='test_middleware.simple_view_prefixed'),
    url(r'^test_middleware/simple_view_template_response/$', test_middleware.simple_view, name='test_middleware.simple_view_template_response_prefixed'),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function  # noqa: FI14

import os
import re
import subprocess
import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

tests_requires = [
    'coverage==4.4.1',
    'mock>=1.0.1',
    'flake8>=2.1.0',
    'codecov>=2.0.0',
    'flake8-future-import>=0.4.5',
    'django-functest==1.0.4',
    'check-manifest',
]


class MyTestCommand(TestCommand):

    description = 'run linters, tests and create a coverage report'
    user_options = []

    def run_tests(self):
        self._run(['flake8', 'django_ftl', 'test'])
        self._run(['./runtests.py'])
        self._run(['check-manifest'])

    def _run(self, command):
        try:
            subprocess.check_call(command)
        except subprocess.CalledProcessError as error:
            print('Command "{0}" failed with exit code'.format(" ".join(command)), error.returncode)
            sys.exit(error.returncode)


def get_version(*file_paths):
    """Retrieves the version from django_ftl/__init__.py"""
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


version = get_version("src", "django_ftl", "__init__.py")


if sys.argv[-1] == 'publish':
    try:
        import wheel
        print("Wheel version: ", wheel.__version__)
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    sys.exit()

if sys.argv[-1] == 'tag':
    print("Tagging the version on git:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

readme = open('README.rst').read()
if hasattr(readme, 'decode'):
    readme = readme.decode('utf-8')
history = open('HISTORY.rst').read()
if hasattr(history, 'decode'):
    history = history.decode('utf-8')
history = history.replace(u'.. :changelog:', u'')

setup(
    name='django-ftl',
    version=version,
    description="""Django bindings for 'fluent', the localization system for today's world.""",
    long_description=readme + u'\n\n' + history,
    author='Luke Plant',
    author_email='L.Plant.98@cantab.net',
    url='https://github.com/django-ftl/django-ftl',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=[
        'fluent.runtime',
        'Django>=1.11',
    ],
    tests_require=tests_requires,  # for 'setup.py test'
    extras_require={
        'develop': tests_requires,  # for 'pip install fluent.runtime[develop]'
    },
    license="MIT",
    zip_safe=False,
    keywords='django-ftl',
    cmdclass={'test': MyTestCommand},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, unicode_literals

from django_ftl import activate
from django_ftl.bundles import Bundle

from .base import TestBase

ftl_bundle = Bundle(['tests/docs.ftl'])


class TestDocs(TestBase):

    def test_usage_docs(self):
        # These tests parallel the code in the usage.rst docs
        activate('en')
        title = ftl_bundle.format('events-title')
        self.assertEqual(title, 'MyApp Events!')

        greeting = ftl_bundle.format('events-greeting', {'username': 'boaty mcboatface'})
        self.assertEqual(greeting, 'Hello, \u2068boaty mcboatface\u2069')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django_ftl.bundles import Bundle

# This module should raise an exception if you try to import it


main = Bundle(['tests/main.ftl'],
              default_locale='en',
              require_activate=True)


class MyThing(object):
    my_label = main.format('simple')
